@@ -4,6 +4,7 @@ This can be used for creation of Python REPLs.
 """
 import __future__
 
+import threading
 from asyncio import get_event_loop
 from functools import partial
 from typing import TYPE_CHECKING, Any, Callable, Dict, Generic, List, Optional, TypeVar
@@ -42,7 +43,7 @@ from prompt_toolkit.key_binding.bindings.open_in_editor import (
     load_open_in_editor_bindings,
 )
 from prompt_toolkit.key_binding.vi_state import InputMode
-from prompt_toolkit.lexers import DynamicLexer, Lexer, PygmentsLexer, SimpleLexer
+from prompt_toolkit.lexers import DynamicLexer, Lexer, SimpleLexer
 from prompt_toolkit.output import ColorDepth, Output
 from prompt_toolkit.styles import (
     AdjustBrightnessStyleTransformation,
@@ -54,7 +55,6 @@ from prompt_toolkit.styles import (
 )
 from prompt_toolkit.utils import is_windows
 from prompt_toolkit.validation import ConditionalValidator, Validator
-from pygments.lexers import Python3Lexer as PythonLexer
 
 from .completer import CompletePrivateAttributes, HidePrivateCompleter, PythonCompleter
 from .history_browser import PythonHistory
@@ -64,9 +64,11 @@ from .key_bindings import (
     load_sidebar_bindings,
 )
 from .layout import CompletionVisualisation, PtPythonLayout
+from .lexer import PtpythonLexer
 from .prompt_style import ClassicPrompt, IPythonPrompt, PromptStyle
+from .signatures import Signature, get_signatures_using_eval, get_signatures_using_jedi
 from .style import generate_style, get_all_code_styles, get_all_ui_styles
-from .utils import get_jedi_script_from_document
+from .utils import unindent_code
 from .validator import PythonValidator
 
 __all__ = ["PythonInput"]
@@ -210,7 +212,7 @@ class PythonInput:
             lambda: self.complete_private_attributes,
         )
         self._validator = _validator or PythonValidator(self.get_compiler_flags)
-        self._lexer = _lexer or PygmentsLexer(PythonLexer)
+        self._lexer = PtpythonLexer(_lexer)
 
         self.history: History
         if history_filename:
@@ -259,7 +261,7 @@ class PythonInput:
 
         self.enable_syntax_highlighting: bool = True
         self.enable_fuzzy_completion: bool = False
-        self.enable_dictionary_completion: bool = False
+        self.enable_dictionary_completion: bool = False  # Also eval-based completion.
         self.complete_private_attributes: CompletePrivateAttributes = (
             CompletePrivateAttributes.ALWAYS
         )
@@ -329,7 +331,7 @@ class PythonInput:
         self.current_statement_index: int = 1
 
         # Code signatures. (This is set asynchronously after a timeout.)
-        self.signatures: List[Any] = []
+        self.signatures: List[Signature] = []
 
         # Boolean indicating whether we have a signatures thread running.
         # (Never run more than one at the same time.)
@@ -916,36 +918,16 @@ class PythonInput:
         loop = loop or get_event_loop()
 
         def run():
-            script = get_jedi_script_from_document(
+            # First, get signatures from Jedi. If we didn't found any and if
+            # "dictionary completion" (eval-based completion) is enabled, then
+            # get signatures using eval.
+            signatures = get_signatures_using_jedi(
                 document, self.get_locals(), self.get_globals()
             )
-
-            # Show signatures in help text.
-            if script:
-                try:
-                    signatures = script.get_signatures()
-                except ValueError:
-                    # e.g. in case of an invalid \\x escape.
-                    signatures = []
-                except Exception:
-                    # Sometimes we still get an exception (TypeError), because
-                    # of probably bugs in jedi. We can silence them.
-                    # See: https://github.com/davidhalter/jedi/issues/492
-                    signatures = []
-                else:
-                    # Try to access the params attribute just once. For Jedi
-                    # signatures containing the keyword-only argument star,
-                    # this will crash when retrieving it the first time with
-                    # AttributeError. Every following time it works.
-                    # See: https://github.com/jonathanslenders/ptpython/issues/47
-                    #      https://github.com/davidhalter/jedi/issues/598
-                    try:
-                        if signatures:
-                            signatures[0].params
-                    except AttributeError:
-                        pass
-            else:
-                signatures = []
+            if not signatures and self.enable_dictionary_completion:
+                signatures = get_signatures_using_eval(
+                    document, self.get_locals(), self.get_globals()
+                )
 
             self._get_signatures_thread_running = False
 
@@ -956,11 +938,8 @@ class PythonInput:
 
                 # Set docstring in docstring buffer.
                 if signatures:
-                    string = signatures[0].docstring()
-                    if not isinstance(string, str):
-                        string = string.decode("utf-8")
                     self.docstring_buffer.reset(
-                        document=Document(string, cursor_position=0)
+                        document=Document(signatures[0].docstring, cursor_position=0)
                     )
                 else:
                     self.docstring_buffer.reset()
@@ -996,3 +975,67 @@ class PythonInput:
                 app.vi_state.input_mode = InputMode.INSERT
 
         asyncio.ensure_future(do_in_terminal())
+
+    def read(self) -> str:
+        """
+        Read the input.
+
+        This will run the Python input user interface in another thread, wait
+        for input to be accepted and return that. By running the UI in another
+        thread, we avoid issues regarding possibly nested event loops.
+
+        This can raise EOFError, when Control-D is pressed.
+        """
+        # Capture the current input_mode in order to restore it after reset,
+        # for ViState.reset() sets it to InputMode.INSERT unconditionally and
+        # doesn't accept any arguments.
+        def pre_run(
+            last_input_mode: InputMode = self.app.vi_state.input_mode,
+        ) -> None:
+            if self.vi_keep_last_used_mode:
+                self.app.vi_state.input_mode = last_input_mode
+
+            if not self.vi_keep_last_used_mode and self.vi_start_in_navigation_mode:
+                self.app.vi_state.input_mode = InputMode.NAVIGATION
+
+        # Run the UI.
+        result: str = ""
+        exception: Optional[BaseException] = None
+
+        def in_thread() -> None:
+            nonlocal result, exception
+            try:
+                while True:
+                    try:
+                        result = self.app.run(pre_run=pre_run)
+
+                        if result.lstrip().startswith("\x1a"):
+                            # When the input starts with Ctrl-Z, quit the REPL.
+                            # (Important for Windows users.)
+                            raise EOFError
+
+                        # Remove leading whitespace.
+                        # (Users can add extra indentation, which happens for
+                        # instance because of copy/pasting code.)
+                        result = unindent_code(result)
+
+                        if result and not result.isspace():
+                            return
+                    except KeyboardInterrupt:
+                        # Abort - try again.
+                        self.default_buffer.document = Document()
+                    except BaseException as e:
+                        exception = e
+                        return
+
+            finally:
+                if self.insert_blank_line_after_input:
+                    self.app.output.write("\n")
+
+        thread = threading.Thread(target=in_thread)
+        thread.start()
+        thread.join()
+
+        if exception is not None:
+            raise exception
+        return result
